'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals
from webuntis import utils, objects, errors

try:
    # Python 3
    import urllib.request as urlrequest
    import urllib.error as urlerrors
except ImportError:
    # Python 2
    import urllib2
    urlrequest = urlerrors = urllib2
import logging
import datetime


try:
    import json  # Python >= 2.6
except ImportError:
    import simplejson as json  # from dependency "simplejson"


class JSONRPCRequest(object):
    _errorcodes = {
        -32601: errors.MethodNotFoundError,
        -8504: errors.BadCredentialsError,
        -8520: errors.NotLoggedInError
    }
    '''This lists the API-errorcodes python-webuntis is able to interpret,
    together with the exception that will be thrown.'''

    def __init__(self, session, method, params=None):
        self._session = session
        self._method = method
        self._params = params or {}

    def request(self):
        '''
        A method for sending a JSON-RPC request.

        :param method: The JSON-RPC method to be executed
        :type method: str

        :param params: JSON-RPC parameters to the method (should be JSON
        serializable)
        :type params: dict
        '''

        url = self._session.options['server'] + \
            '?school=' + \
            self._session.options['school']

        headers = {
            'User-Agent': self._session.options['useragent'],
            'Content-Type': 'application/json'
        }

        request_body = {
            'id': str(datetime.datetime.today()),
            'method': self._method,
            'params': self._params,
            'jsonrpc': '2.0'
        }

        if self._method != 'authenticate':
            if 'jsessionid' not in self._session.options:
                raise errors.NotLoggedInError(
                    'Don\'t have JSESSIONID. Did you already log out?')
            else:
                headers['Cookie'] = 'JSESSIONID=' + \
                    self._session.options['jsessionid']

        logging.debug('Making new request:')
        logging.debug('URL: ' + url)
        logging.debug('DATA: ' + str(request_body))

        result_body = self._send_request(
            url,
            json.dumps(request_body).encode(),
            headers
        )
        return self._parse_result(request_body, result_body)

    def _parse_result(self, request_body, result_body):
        '''A subfunction of _make_request that, given the decoded JSON result,
        handles the error codes or, if everything went well, returns the result
        attribute of it. The request data has to be given too for logging and
        ID validation.

        :param request_body: The not-yet-encoded body of the request sent.
        :param result_body: The decoded body of the result recieved.
        '''

        if request_body['id'] != result_body['id']:
            raise errors.RemoteError(
                'Request ID was not the same one as returned.')

        try:
            return result_body['result']
        except KeyError:
            self._parse_error_code(request_body, result_body)

    def _parse_error_code(self, request_body, result_body):
        '''A helper function for handling JSON error codes.'''
        logging.error(result_body)
        try:
            error = result_body['error']
            exc = self._errorcodes[error['code']](error['message'])
        except KeyError:
            exc = errors.RemoteError(
                ('Some JSON-RPC-ish error happened. Please report this to the '
                    'developer so he can implement a proper handling.'),
                str(result_body),
                str(request_body)
            )

        raise exc

    def _send_request(self, url, data, headers):
        '''A subfunction of _make_request, mostly because of mocking. Sends the
        given headers and data to the url and tries to return the result
        JSON-decoded.
        '''

        request = urlrequest.Request(
            url,
            data,
            headers
        )
        # this will eventually raise errors, e.g. if there's an unexpected http
        # status code
        result_obj = urlrequest.urlopen(request)
        result = result_obj.read().decode('utf-8')

        try:
            result_data = json.loads(result)
            logging.debug('Valid JSON found')
            logging.debug(result_data)
        except ValueError:
            raise errors.RemoteError('Invalid JSON', str(result))
        else:
            return result_data


class JSONRPCSession(object):
    '''Lower-level version of :py:class:`Session`. Do not use this.'''

    options = None
    '''Contains a options dict upon initialization. See
    :py:class:`webuntis.utils.option_utils` for more information.
    '''

    def __init__(self, **kwargs):
        # The OptionStore is an extended dictionary, associating validators
        # and other helper methods with each key
        self.options = utils.FilterDict(utils.option_utils.options)
        options = {
            'server': None,
            'school': None,
            'useragent': None,
            'username': None,
            'password': None,
            'jsessionid': None,
            'login_repeat': 0
        }
        options.update(kwargs)
        self.options.update(options)

    def __enter__(self):
        '''Context-manager'''
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''Context-manager -- the only thing we need to clean up is to log out
        '''
        self.logout(suppress_errors=True)

    def logout(self, suppress_errors=False):
        '''
        Log out of session

        :param suppress_errors: boolean, whether to suppress errors if we
            already were logged out.

        :raises: :py:class:`webuntis.errors.NotLoggedInError`
        '''
        def throw_errors():
            if not suppress_errors:
                raise errors.NotLoggedInError('We already were logged out.')

        try:
            # Send a JSON-RPC 'logout' method without parameters to log out
            self._request('logout')
        except errors.NotLoggedInError:
            throw_errors()

        try:
            del self.options['jsessionid']
        except KeyError:
            throw_errors()

    def login(self):
        '''Initializes an authentication, provided we have the credentials for
        it.

        :returns: The session. This is useful for jQuery-like command
            chaining::

                s = webuntis.Session(...).login()

        :raises: :py:class:`webuntis.errors.BadCredentialsError`
        :raises: :py:class:`webuntis.errors.AuthError`
        '''

        if 'username' not in self.options \
                or 'password' not in self.options:
            raise errors.AuthError('No login data specified.')

        logging.debug('Trying to authenticate with username/password...')
        logging.debug('Username: ' +
                      self.options['username'] +
                      ' Password: ' +
                      self.options['password'])
        res = self._request('authenticate', {
            'user': self.options['username'],
            'password': self.options['password'],
            'client': self.options['useragent']
        }, use_login_repeat=False)
        logging.debug(res)
        if 'sessionId' in res:
            logging.debug('Did get a jsessionid from the server:')
            self.options['jsessionid'] = res['sessionId']
            logging.debug(self.options['jsessionid'])
        else:
            raise errors.AuthError(
                'Something went wrong while authenticating',
                res
            )

        return self

    def _request(self, method, params=None, use_login_repeat=None):
        if use_login_repeat is None:
            use_login_repeat = (method not in ('logout', 'authenticate'))
        attempts_left = self.options['login_repeat'] if use_login_repeat else 0

        data = None

        while data is None:
            try:
                data = JSONRPCRequest(self, method, params).request()
            except errors.NotLoggedInError:
                if attempts_left > 0:
                    self.logout(suppress_errors=True)
                    self.login()
                else:
                    raise errors.NotLoggedInError(
                        'Tried to login several times, failed. Original method '
                        'was ' + method)
            else:
                return data

            attempts_left -= 1  # new round!


class Session(JSONRPCSession):
    '''The origin of everything you want to do with the WebUntis API. Can be
    used as a context-handler.'''

    _cache = None
    '''Contains the caching dictionary for requests.'''

    def __init__(self, **options):
        try:
            cachelen = options['cachelen']
            del options['cachelen']
        except KeyError:
            cachelen = 20

        self._cache = utils.LruDict(maxlen=cachelen)

        JSONRPCSession.__init__(self, **options)

    def _make_cache_key(self, method, kwargs):
        '''A helper method that generates a hashable object out of a string and
        a dictionary.

        It doesn't use ``hash()`` or similar methods because it's neat that the
        keys are human-readable and enable us to trace back the origin of the
        key. Python does that anyway under the hood when using it as a
        dictionary key.
        '''

        return (method, frozenset((kwargs or {}).items()))

    def __getattr__(self, name):
        '''Returns a callable which creates an instance (or reuses an old one)
        of the appropriate object-list class
        '''
        def result_object_wrapper(**kwargs):
            key = self._make_cache_key(name, kwargs)

            def get_result_object():
                obj = objects.result_objects[name](session=self, kwargs=kwargs)
                obj.get_data()
                return obj

            if key not in self._cache:
                obj = self._cache[key] = get_result_object()
            else:
                obj = self._cache[key]
            return obj

        if name in objects.result_objects:
            return result_object_wrapper
        else:
            raise AttributeError(name)



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals
import itertools


def table(periods, width=None):
    '''
    Creates a table-like nested list out of a list of periods.

    :param timetable: A :py:class:`webuntis.objects.PeriodList` instance or any
    other iterable containing :py:class:`webuntis.objects.PeriodObject`
    instances.

    :param width: Optionally, a fixed width for the table. The function will
    fill every row with empty cells until that width is met. If the timetable
    is too big, it will raise a ``ValueError``.

    :returns: A list containing "rows", which in turn contain "hours", which
    contain :py:class:`webuntis.objects.PeriodObject` instances which are
    happening at the same time.

    Example::

        today = datetime.datetime.today()
        monday = today - datetime.timedelta(days=today.weekday())
        friday = monday + datetime.timedelta(days=4)

        table = s.timetable(klasse=878, start=monday, end=friday) \
                .to_table(width=5)

        print('<table><thead>')
        for weekday in range(5):
            print('<th>' + str(weekday) + '</th>')

        print('</thead><tbody>')
        for time, row in table:
            print('<tr>')
            for weekday_number, cell in row:
                print('<td>')
                for hour in cell:
                    print('<div>')
                    print(', '.join(su.name for su in hour.subjects))
                    print('</div>')

                print('</td>')

            print('</tr>')

        print('</tbody></table>')

    Gives you HTML like this:

    +--------+--------+--------+--------+--------+
    | 0      | 1      | 2      | 3      | 4      |
    +========+========+========+========+========+
    | ME     | M      | PH     | M      | GSK    |
    +--------+--------+--------+--------+--------+
    | M      | BU     | D      | FRA    | D      |
    |        |        |        | LAT    |        |
    |        |        |        | SPA    |        |
    +--------+--------+--------+--------+--------+
    | E      | BU     | FRA    | BU     | E      |
    |        |        | LAT    |        |        |
    |        |        | SPA    |        |        |
    +--------+--------+--------+--------+--------+
    | RK     | GSK    | E      | ME     | GWK    |
    | RISL   |        |        |        |        |
    +--------+--------+--------+--------+--------+
    | D      | BE     | M      | PH     | PH     |
    +--------+--------+--------+--------+--------+
    | FRA    |        |        |        |        |
    | LAT    |        |        |        |        |
    | SPA    |        |        |        |        |
    +--------+--------+--------+--------+--------+
    | INF+   |        |        |        |        |
    +--------+--------+--------+--------+--------+
    | INF+   |        |        |        |        |
    +--------+--------+--------+--------+--------+
    | BSP    | RKO    |        |        |        |
    +--------+--------+--------+--------+--------+

    '''

    if not len(periods):
        return []

    time = lambda x: int(x.start.strftime('%H%M%S'))
    date = lambda x: int(x.start.strftime('%Y%m%d'))

    grouped = []
    times = []
    for i, row in itertools.groupby(sorted(periods, key=time), time):
        new_row = []

        starttime = None

        # group lessons by date -- now all lessons occuring at the same time
        # are grouped
        for i2, hour in itertools.groupby(sorted(row, key=date), date):
            if hour:
                # itertools._grouper objects can't be accessed like lists
                hour = list(hour)
                weekday = hour[0].start.weekday()

                while (len(new_row) - 1) < weekday:
                    new_row.append(list())  # expand row as much as needed
                new_row[weekday] = hour

                if not starttime:
                    starttime = hour[0].start

        grouped.append(new_row)
        times.append(starttime)

    # expand each row to the maximal length
    longest_row = len(max(grouped, key=len))

    if width and longest_row > width:
        raise ValueError(
            'Fixed width too small. Need at least %i' % longest_row)

    longest_row = width

    for row in grouped:
        while len(row) < longest_row:
            row.append(list())

    # at last add a weekday indicator
    return zip(times, [enumerate(row) for row in grouped])



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''


from __future__ import unicode_literals

import unittest
import mock
import os
import json
import webuntis


tests_path = os.path.abspath(os.path.dirname(__file__))
data_path = tests_path + '/static'


def get_json_resource(name):
        with open(os.path.join(data_path, name)) as f:
            return json.load(f)


class TestCaseBase(unittest.TestCase):
    pass


class OfflineTestCase(TestCaseBase):
    def setUp(self):
        class methods(object):
            '''This eliminates any remote access.'''
            pass

        self.request_patcher = patcher = mock_results(methods)
        patcher.start()

        init_params = {
            'useragent': 'fooagent',
            'school': 'fooschool',
            'username': 'foouser',
            'password': 'hunter2',
            'server': 'fooboo_server',
            'jsessionid': 'CHUCK_TESTA'
        }

        self._session_init_params = init_params

        self.session = webuntis.Session(**init_params)

    def tearDown(self):
        self.request_patcher.stop()
        self.session = None


def mock_results(methods, use_original=False):
    def callback(self, url, data, headers):
        jsondata = json.loads(data.decode('utf-8'))
        method = jsondata['method']

        data = getattr(methods, method)(self, url, jsondata, headers)
        d = {'id': jsondata['id']}
        d.update(data)
        return d

    return mock.patch(
        'webuntis.session.JSONRPCRequest._send_request',
        new=callback
    )



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''
from __future__ import unicode_literals

import mock
import webuntis
import webuntis.utils as utils
import datetime

from webuntis.tests.utils import OfflineTestCase, mock_results, \
    get_json_resource


class BasicUsageTests(OfflineTestCase):
    '''Mock the _get_data method of all result objects and test the package
    against basic usage.'''

    def test_getdepartments_mock(self):
        jsonstr = get_json_resource('getdepartments_mock.json')

        class methods(object):
            @staticmethod
            def getDepartments(self, url, jsondata, headers):
                return {'result': jsonstr}

        with mock_results(methods):
            for dep_raw, dep in zip(jsonstr, self.session.departments()):
                self.assertEqual(dep_raw['id'], dep.id)
                self.assertEqual(dep_raw['longName'], dep.long_name)
                self.assertEqual(dep_raw['name'], dep.name)

    def test_getholidays_mock(self):
        jsonstr = get_json_resource('getholidays_mock.json')

        class methods(object):
            @staticmethod
            def getHolidays(self, url, jsondata, headers):
                return {'result': jsonstr}

        with mock_results(methods):
            for holiday_raw, holiday in zip(jsonstr, self.session.holidays()):
                self.assertEqual(holiday_raw['id'], holiday.id)
                self.assertEqual(holiday_raw['name'], holiday.short_name)
                self.assertEqual(holiday_raw['longName'], holiday.name)

                self.assertEqual(
                    holiday_raw['startDate'],
                    int(holiday.start.strftime('%Y%m%d'))
                )
                self.assertEqual(
                    holiday_raw['endDate'],
                    int(holiday.end.strftime('%Y%m%d'))
                )

    def test_getklassen_mock(self):
        jsonstr = get_json_resource('getklassen_mock.json')

        class methods(object):
            @staticmethod
            def getKlassen(self, url, jsondata, headers):
                return {'result': jsonstr}

        with mock_results(methods):
            klassen = self.session.klassen()
            for klasse_raw, klasse in zip(jsonstr, klassen):
                self.assertEqual(klasse_raw['id'], klasse.id)
                self.assertEqual(klasse_raw['name'], klasse.name)
                self.assertEqual(klasse_raw['longName'], klasse.long_name)

            self.assertEqual(klassen.filter(id=129)[0].id, 129)
            self.assertEqual(
                [129, 130, 137],
                sorted(
                    kl.id for kl in
                    klassen.filter(id=set([129, 130, 137]))
                )
            )

            self.assertTrue({'id': 129} in klassen)
            self.assertTrue({'name': '6A'} in klassen)

    def test_gettimetables_mock(self):
        jsonstr = get_json_resource('gettimetables_mock.json')
        jsonstr_kl = get_json_resource('getklassen_mock.json')
        jsonstr_te = get_json_resource('getteachers_mock.json')
        jsonstr_su = get_json_resource('getsubjects_mock.json')
        jsonstr_ro = get_json_resource('getrooms_mock.json')

        class methods(object):
            @staticmethod
            def getTimetable(self, url, jsondata, headers):
                return {'result': jsonstr}

            @staticmethod
            def getKlassen(self, url, jsondata, headers):
                return {'result': jsonstr_kl}

            @staticmethod
            def getTeachers(self, url, jsondata, headers):
                return {'result': jsonstr_te}

            @staticmethod
            def getSubjects(self, url, jsondata, headers):
                return {'result': jsonstr_su}

            @staticmethod
            def getRooms(self, url, jsondata, headers):
                return {'result': jsonstr_ro}

        with mock_results(methods):
            tt = self.session.timetable(klasse=114)

            for period_raw, period in zip(jsonstr, tt):
                self.assertEqual(
                    int(period.start.strftime('%H%M')),
                    period_raw['startTime']
                )
                self.assertEqual(
                    int(period.start.strftime('%Y%m%d')),
                    period_raw['date']
                )

                self.assertEqual(
                    int(period.end.strftime('%H%M')),
                    period_raw['endTime']
                )
                self.assertEqual(
                    int(period.end.strftime('%H%M')),
                    period_raw['endTime']
                )
                if 'code' in period_raw:
                    self.assertEqual(period_raw['code'], period.code)
                else:
                    self.assertEqual(period.code, None)

                if 'type' in period_raw:
                    self.assertEqual(period_raw['lstype'], period.type)
                else:
                    self.assertEqual(period.type, 'ls')

                self.assertEqual(len(period_raw['kl']), len(period.klassen))
                for klasse_raw, klasse in zip(period_raw['kl'], period.klassen):
                    self.assertEqual(klasse.id, klasse_raw['id'])

                self.assertEqual(len(period_raw['te']), len(period.teachers))
                for teacher_raw, teacher in zip(period_raw['te'], period.teachers):
                    self.assertEqual(teacher.id, teacher_raw['id'])

                self.assertEqual(len(period_raw['su']), len(period.subjects))
                for subject_raw, subject in zip(period_raw['su'], period.subjects):
                    self.assertEqual(subject.id, subject_raw['id'])

                self.assertEqual(len(period_raw['ro']), len(period.rooms))
                for room_raw, room in zip(period_raw['ro'], period.rooms):
                    self.assertEqual(room.id, room_raw['id'])

    def test_getrooms_mock(self):
        jsonstr = get_json_resource('getrooms_mock.json')

        class methods(object):
            @staticmethod
            def getRooms(self, url, jsondata, headers):
                return {'result': jsonstr}

        with mock_results(methods):
            for room_raw, room in zip(jsonstr, self.session.rooms()):
                self.assertEqual(room_raw['longName'], room.long_name)
                self.assertEqual(room_raw['name'], room.name)
                self.assertEqual(room_raw['id'], room.id)

    def test_getschoolyears_mock(self):
        jsonstr = get_json_resource('getschoolyears_mock.json')
        current_json = jsonstr[3]

        class methods(object):
            @staticmethod
            def getSchoolyears(self, url, jsondata, headers):
                return {'result': jsonstr}

            @staticmethod
            def getCurrentSchoolyear(self, url, jsondata, headers):
                return {'result': current_json}

        with mock_results(methods):
            schoolyears = self.session.schoolyears()

            self.assertEqual(current_json['id'], schoolyears.current.id)
            self.assertTrue(schoolyears.current.is_current)
            current_count = 0
            for year_raw, year in zip(jsonstr, schoolyears):
                self.assertEqual(year_raw['id'], year.id)
                self.assertEqual(year_raw['name'], year.name)

                self.assertEqual(
                    year_raw['startDate'],
                    int(year.start.strftime('%Y%m%d'))
                )
                self.assertEqual(
                    year_raw['endDate'],
                    int(year.end.strftime('%Y%m%d'))
                )
                if year.is_current:
                    self.assertEqual(year, schoolyears.current)
                    current_count += 1
                    self.assertEqual(current_count, 1)

    def test_getsubjects_mock(self):
        jsonstr = get_json_resource('getsubjects_mock.json')

        class methods(object):
            @staticmethod
            def getSubjects(self, url, jsondata, headers):
                return {'result': jsonstr}

        with mock_results(methods):
            for subj_raw, subj in zip(jsonstr, self.session.subjects()):
                self.assertEqual(subj_raw['id'], subj.id)
                self.assertEqual(subj_raw['name'], subj.name)
                self.assertEqual(subj_raw['longName'], subj.long_name)

    def test_getteachers_mock(self):
        jsonstr = get_json_resource('getteachers_mock.json')

        class methods(object):
            @staticmethod
            def getTeachers(self, url, jsondata, headers):
                return {'result': jsonstr}

        with mock_results(methods):
            for t_raw, t in zip(jsonstr, self.session.teachers()):
                self.assertEqual(t_raw['longName'], t.long_name)
                self.assertEqual(t_raw['longName'], t.surname)
                self.assertEqual(t_raw['foreName'], t.fore_name)
                self.assertEqual(t_raw['name'], t.name)

    def test_gettimegrid_mock(self):
        jsonstr = get_json_resource('gettimegrid_mock.json')

        class methods(object):
            @staticmethod
            def getTimegridUnits(self, url, jsondata, headers):
                return {'result': jsonstr}

        with mock_results(methods):
            for t_raw, t in zip(jsonstr, self.session.timegrid()):
                self.assertEqual(t_raw['day'], t.day)
                for t2_raw, t2 in zip(t_raw['timeUnits'], t.times):
                    self.assertEqual(t2_raw['startTime'],
                                     int(t2[0].strftime('%H%M')))
                    self.assertEqual(t2_raw['endTime'],
                                     int(t2[1].strftime('%H%M')))

    def test_getstatusdata_mock(self):
        jsonstr = get_json_resource('getstatusdata_mock.json')

        class methods(object):
            @staticmethod
            def getStatusData(self, url, jsondata, headers):
                return {'result': jsonstr}

        def validate_statusdata(raw, processed):
            name = list(raw.items())[0][0]
            colors = raw[name]
            self.assertEqual(name, processed.name)
            self.assertEqual(colors['foreColor'], processed.forecolor)
            self.assertEqual(colors['backColor'], processed.backcolor)

        with mock_results(methods):
            statusdata = self.session.statusdata()
            for lstype_raw, lstype in zip(jsonstr['lstypes'],
                                          statusdata.lesson_types):
                validate_statusdata(lstype_raw, lstype)


            for code_raw, code in zip(jsonstr['codes'], statusdata.period_codes):
                validate_statusdata(code_raw, code)
                

    def test_login_repeat_invalid_session(self):
        retry_amount = 5
        calls = []

        # This produces a list of 5 * ['getCurrentSchoolyear'] with ['logout',
        # 'authenticate'] between each.
        expected_calls = (
            ['getCurrentSchoolyear', 'logout', 'authenticate']
            * (retry_amount + 1)
        )[:-2]

        class methods(object):
            @staticmethod
            def authenticate(self, url, jsondata, headers):
                calls.append(jsondata['method'])
                return {
                    'result': {'sessionId': 'Foobar_session_' + jsondata['id']}
                }

            @staticmethod
            def getCurrentSchoolyear(self, url, jsondata, headers):
                calls.append(jsondata['method'])
                return {
                    'error': {'code': -8520, 'message': 'Not Logged In!'}
                }

            @staticmethod
            def logout(self, url, jsondata, headers):
                calls.append(jsondata['method'])
                return {
                    'result': {'bla': 'blub'}
                }

        with mock_results(methods):
            with mock.patch.dict(
                self.session.options,
                {'login_repeat': retry_amount}
            ):
                self.assertRaises(webuntis.errors.NotLoggedInError,
                                  self.session._request, 'getCurrentSchoolyear')

        self.assertEqual(calls, expected_calls)

    def test_login_repeat_not_logged_in(self):
        retry_amount = 1
        calls = []

        # This produces a list of 5 * ['getCurrentSchoolyear'] with ['logout',
        # 'authenticate'] between each.
        expected_calls = ['authenticate', 'getCurrentSchoolyear']

        class methods(object):
            @staticmethod
            def authenticate(self, url, jsondata, headers):
                calls.append(jsondata['method'])
                return {
                    'id': jsondata['id'],
                    'result': {'sessionId': 'Foobar_session_' + jsondata['id']}
                }

            @staticmethod
            def _nope(self, url, jsondata, headers):
                calls.append(jsondata['method'])
                return {
                    'id': jsondata['id'],
                    'error': {'code': -8520, 'message': 'Not Logged In!'}
                }

            getCurrentSchoolyear = logout = _nope
            del _nope

        with mock_results(methods):
            with mock.patch.dict(
                self.session.options,
                {'login_repeat': retry_amount, 'jsessionid': None}
            ):
                self.assertRaises(webuntis.errors.NotLoggedInError,
                                  self.session._request, 'getCurrentSchoolyear')

        self.assertEqual(calls, expected_calls)


class InternalTests(OfflineTestCase):
    '''Test certain internal interfaces, such as utils'''

    def test_make_cache_key(self):
        # The hash builtin will take care of us if the results aren't hashable.
        hash(self.session._make_cache_key('getStuff', {'foo': 'bar'}))
        hash(self.session._make_cache_key('getStuff', {}))
        hash(self.session._make_cache_key('getStuff', None))

    def test_is_iterable_util(self):
        tests = [
            ((), True),
            (None, False),
            ([], True),
            ({}, True),
            ("FOO", False),
            (str("FOO"), False),
            (123, False)
        ]

        for given_input, expected_output in tests:
            self.assertEqual(utils.is_iterable(given_input), expected_output)

    def test_options_invalidattribute(self):
        self.assertFalse('nigglywiggly' in self.session.options)
        self.assertRaises(
            KeyError,
            self.session.options.__getitem__,
            'nigglywiggly'
        )

    def test_datetime_utils(self):
        obj = utils.datetime_utils.parse_datetime(20121005, 0)
        self.assertEqual(obj.year, 2012)
        self.assertEqual(obj.month, 10)
        self.assertEqual(obj.day, 5)
        self.assertEqual(obj.hour, 0)
        self.assertEqual(obj.minute, 0)
        self.assertEqual(obj.second, 0)

    def test_filterdict(self):
        store = utils.FilterDict({
            'whatever': lambda x: x,
            'always_whoop': lambda x: 'whoop'
        })
        store['whatever'] = 'lel'
        self.assertEqual(store['whatever'], 'lel')

        store['always_whoop'] = 'what'
        self.assertEqual(store['always_whoop'], 'whoop')

        del store['whatever']
        self.assertRaises(KeyError, store.__getitem__, 'whatever')

        del store['always_whoop']
        self.assertRaises(KeyError, store.__getitem__, 'always_whoop')

    def test_session_invalidattribute(self):
        self.assertRaises(AttributeError, getattr, self.session, 'foobar')
        self.assertFalse(hasattr(self.session, 'foobar'))

    def test_requestcaching(self):
        jsonstr = get_json_resource('getklassen_mock.json')

        def result_mock(s, method, params=None, use_login_repeat=None):
            self.assertEqual(method, 'getKlassen')
            return jsonstr

        with mock.patch(
            'webuntis.session.Session._request',
            new=result_mock
        ):
            self.session.klassen()

        with mock.patch(
            'webuntis.session.Session._request',
            side_effect=Exception('CHUCK TESTA')
        ):
            self.session.klassen()

    def test_listitem(self):
        session = self.session
        parent = None
        data = {'id': 42}
        item = webuntis.objects.ListItem(session=session, parent=parent, data=data)

        self.assertEqual(item._session, session)
        self.assertEqual(item._parent, parent)
        self.assertEqual(item._data, data)
        self.assertEqual(item.id, data['id'])
        self.assertEqual(int(item), item.id)

    def test_result_object_without_parameters_method(self):
        with mock.patch.object(
            webuntis.objects.Result,
            '_jsonrpc_parameters',
            new=False
        ):
            obj = webuntis.objects.Result(session=self.session, kwargs={})
            self.assertRaises(NotImplementedError, obj.get_data)

    def test_optionparsers_server(self):
        tests = [
            ('webuntis.grupet.at',
                'http://webuntis.grupet.at/WebUntis/jsonrpc.do'),
            ('https://webuntis.grupet.at',
                'https://webuntis.grupet.at/WebUntis/jsonrpc.do'),
            ('webuntis.grupet.at:8080',
                'http://webuntis.grupet.at:8080/WebUntis/jsonrpc.do'),
            ('webuntis.grupet.at/a/b/c', 'http://webuntis.grupet.at/a/b/c'),
            ('webuntis.grupet.at/', 'http://webuntis.grupet.at/'),
        ]

        for parser_input, expected_output in tests:
            self.assertEqual(
                webuntis.utils.option_utils.server(parser_input),
                expected_output
            )

        self.assertRaises(ValueError,
                          webuntis.utils.option_utils.server, '!"$%')

    def test_resultobject_get_data(self):
        kwargs = {}
        testclass = webuntis.objects.DepartmentList
        testobj = testclass(session=self.session, kwargs=kwargs)

        class methods(object):
            @staticmethod
            def getDepartments(req, url, jsondata, headers):
                self.assertEqual(jsondata['params'],
                                 testobj._jsonrpc_parameters(**kwargs))
                return {'result': {}}

        with mock_results(methods):
            testobj.get_data()

    def helper_jsonrpc_parameters(self, resultclass):
        '''A wrapper for the _jsonrpc_parameters method of any result-type
        class.'''
        def jsonrpc_parameters(kwargs):
            return resultclass(
                session=self.session,
                kwargs={}
            )._jsonrpc_parameters(
                **kwargs
            )

        return jsonrpc_parameters

    def test_objects_klassenlist_jsonrpc_parameters(self):
        tests = [
            ({'schoolyear': 13}, {'schoolyearId': 13}),
            ({'schoolyear': "123"}, {'schoolyearId': 123})
        ]

        for given_input, expected_output in tests:
            self.assertEqual(
                self.helper_jsonrpc_parameters(webuntis.objects.KlassenList)(given_input),
                expected_output
            )

    def test_objects_periodlist_jsonrpc_parameters(self):
        parambuilder = self.helper_jsonrpc_parameters(webuntis.objects.PeriodList)
        dtobj = datetime.datetime.now()
        dtobj_formatted = int(dtobj.strftime('%Y%m%d'))
        tests = [
            ({'start': None, 'end': None, 'klasse': 123},
                {'id': 123, 'type': 1}),
            ({'start': None, 'end': None, 'teacher': 124},
                {'id': 124, 'type': 2}),
            ({'start': None, 'end': None, 'subject': 154},
                {'id': 154, 'type': 3}),
            ({'start': None, 'end': dtobj, 'subject': 1337},
                {
                    'id': 1337,
                    'type': 3,
                    'startDate': dtobj_formatted,
                    'endDate': dtobj_formatted
                }),
            ({'end': None, 'start': dtobj, 'subject': 1337},
                {
                    'id': 1337,
                    'type': 3,
                    'startDate': dtobj_formatted,
                    'endDate': dtobj_formatted
                })
        ]
        for given_input, expected_output in tests:
            self.assertEqual(
                parambuilder(given_input),
                expected_output
            )

        self.assertRaises(TypeError, parambuilder, {})
        self.assertRaises(TypeError, parambuilder, {'foobar': 123})

    def test_resultclass_invalid_arguments(self):
        self.assertRaises(TypeError, webuntis.objects.Result, session=self.session, kwargs={}, data="LELELE")
        self.assertRaises(TypeError, webuntis.objects.Result)
        self.assertRaises(TypeError, webuntis.objects.Result, session=self.session)



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals
from datetime import datetime as datetime


forms = {
    'date': ('%Y%m%d', 6),
    'time': ('%H%M', 4),
}


def parse_date(string):
    return _parse(string, 'date')


def parse_time(string):
    return _parse(string, 'time')


def parse_datetime(date, time):
    pdate = parse_date(date)
    ptime = parse_time(time)

    newdate = datetime(
        year=pdate.year,
        month=pdate.month,
        day=pdate.day,
        hour=ptime.hour,
        minute=ptime.minute
    )

    return newdate


def _parse(string, form):
    if type(string) == int:
        string = str(string)

    formlen = forms[form][1]
    while len(string) < formlen:
        string = '0' + string

    return datetime.strptime(string, forms[form][0])


def format_date(obj):
    return _format(obj, 'date')


def format_time(obj):
    return _format(obj, 'time')


def _format(obj, form):
    return int(datetime.strftime(obj, forms[form][0]))



'''
    This file is part of python-webuntis

    :copyright: (c) 2012 by Markus Unterwaditzer.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import unicode_literals


class Error(Exception):
    '''Superclass for all `python-webuntis`-specific errors, never gets raised
    directly. Positional arguments get joined with ``", "`` and get dumped into
    ``self.msg``'''

    def __init__(self, *args):
        Exception.__init__(self, *args)
        self.msg = ', '.join(args)


class RemoteError(Error, IOError):
    '''There was some kind of error while interacting with the server.'''
    pass


class MethodNotFoundError(RemoteError):
    '''The JSON-RPC method was not found. This really should not occur.'''
    pass


class AuthError(RemoteError):
    '''Usually missing credentials, but also other problems while logging in
    are covered with this.'''
    pass


class BadCredentialsError(AuthError):
    '''Invalid username or password'''
    pass


class NotLoggedInError(AuthError):
    '''The session expired or we never logged in.'''
